appliances_dict = {}
mqtt_appliances_dict = {}
ack_events = {}
loggers = {}


def get_appliance(device_id: tuple, channel: tuple, appliances_dict: dict):
    # appliance key form is ((device_id),(channels))
    try:
        device_appliances = [
            a
            for k, a in appliances_dict.items()
            if k[0] == device_id and channel in k[1]
        ]
        return tuple(device_appliances)[0]
    except IndexError:
        print(f"No appliances found for key {(device_id,channel)}")
        return None
    except Exception as e:
        print(f"An error occurred: {e}")
        return None



import binascii
from TISControlProtocol.crc import checkCRC, packCRC


def bytes2hex(data, rtype=[]):
    """A helper function to parse bytes to hex

    :param data: the raw bytes array
    :type data: bytes array
    :param rtype: determine return type whether list of ints or single hex str, defaults to []
    :type rtype: list, optional
    :return: list of ints or hex string
    :rtype: list | str
    """
    hex_string = binascii.hexlify(data).decode()
    hex_list = [int(hex_string[i : i + 2], 16) for i in range(0, len(hex_string), 2)]
    if isinstance(rtype, list):
        return hex_list
    else:
        return hex_string


def build_packet(
    operation_code: list,
    ip_address: str,
    destination_mac: str,
    source_mac: str = "CB:CB:CB:CB:CB:CB:CB:CB",
    device_id: list = [],
    source_device_id: list = [0x01, 0xFE],
    additional_packets: list = [],
    header="SMARTCLOUD",
):

    # test if all params are loaded
    ip_bytes = [int(part) for part in ip_address.split(".")]
    header_bytes = [ord(char) for char in header]
    source_mac = [int(part, 16) for part in source_mac.split(":")]
    destination_mac = [int(part, 16) for part in destination_mac.split(":")]
    # TODO: modify length
    # length = 21 + len(additional_packets)
    length = 11 + len(additional_packets)
    packet = (
        ip_bytes
        + header_bytes
        + [0xAA, 0xAA]
        + [length]
        # + source_mac
        + source_device_id
        + [0xFF, 0xFE]
        + operation_code
        # + destination_mac
        + device_id
        + additional_packets
    )
    packet = packCRC(packet)
    return packet


def decode_mac(mac: list):
    return ":".join([f"{byte:02X}" for byte in mac])


def int_to_8_bit_binary(number):
    binary_string = bin(number)[2:]
    return binary_string.zfill(8)[::-1]



from collections import defaultdict


def _parse_device_manager_request(data: dict) -> None:
    """Parse the device manager request."""
    converted = {
        appliance: {
            "device_id": [int(n) for n in details[0]["device_id"].split(",")],
            "appliance_type": details[0]["appliance_type"].lower().replace(" ", "_"),
            "appliance_class": details[0].get("appliance_class", None),
            "is_protected": bool(int(details[0]["is_protected"])),
            "channels": [
                {
                    "channel_number": int(detail["channel_number"]),
                    "channel_type": detail["channel_type"],
                    "channel_name": detail["channel_name"],
                }
                for detail in details
            ],
        }
        for appliance, details in data["appliances"].items()
        if details[0]["gateway"] == "192.168.1.201"
    }

    grouped = defaultdict(list)
    for appliance, details in converted.items():
        if details["appliance_type"] in ["Switch", "Light"]:
            # logging.error(f"appliance: {appliance}, details: {details}")
            grouped[details["appliance_type"]].append({appliance: details})

    _config_entries = dict(grouped)
    # # add a lock module config entry
    # self._config_entries["lock_module"] = {
    #     "password": data["configs"]["lock_module_password"]
    # }
    # logging.error(f"config_entries stored: {self._config_entries}")
    # # await self.update_entities()
    return _config_entries


test_request = {
    "appliances": {
        "appliance1": [
            {
                "device_id": "1,2",
                "appliance_type": "Light",
                "appliance_class": "classA",
                "is_protected": "1",
                "gateway": "192.168.1.201",
                "channels": [
                    {
                        "channel_number": "1",
                        "channel_type": "type1",
                        "channel_name": "Channel 1",
                    }
                ],
            }
        ]
    },
    "configs": {"lock_module_password": "your_password_here"},
}

out = _parse_device_manager_request(test_request)
print(out)



from homeassistant.core import HomeAssistant  # type: ignore
import logging


async def handle_auto_binary_feedback(hass: HomeAssistant, info: dict):
    logging.error(f"Auto Binary Feedback: {info}")
    channels_number: int = info["additional_bytes"][0]
    channels_values: list = info["additional_bytes"][-1 * channels_number :]

    event_data = {
        "device_id": info["device_id"],
        "feedback_type": "auto_binary_feedback",
        "channels_values": channels_values,
    }
    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
    except Exception as e:
        logging.error(f"error in firing event: {e}")



from homeassistant.core import HomeAssistant  # type: ignore
import logging


async def handle_binary_feedback(hass: HomeAssistant, info: dict):
    # remove auxilary bytes which represents number of scenarios
    len_aux = info["additional_bytes"][0]
    info["additional_bytes"] = info["additional_bytes"][len_aux + 1 :]
    event_data = {
        "device_id": info["device_id"],
        "feedback_type": "binary_feedback",
        "additional_bytes": info["additional_bytes"],
    }
    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
    except Exception as e:
        logging.error(f"error in firing event: {e}")



from homeassistant.core import HomeAssistant

import logging
import asyncio
from TISControlProtocol.shared import ack_events


async def handle_control_response(hass: HomeAssistant, info: dict):
    channel_number = info["additional_bytes"][0]
    # await target_appliance.handle_packet(info["additional_bytes"], "control")
    event_data = {
        "device_id": info["device_id"],
        "channel_number": channel_number,
        "feedback_type": "control_response",
        "additional_bytes": info["additional_bytes"],
    }
    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
    except Exception as e:
        logging.error(f"error in firing even for feedbackt: {e}")

    try:
        event: asyncio.Event = ack_events.get(
            (
                tuple(info["device_id"]),
                (0x00, 0x31),
                channel_number,
            )
        )
        if event is not None:
            print("setting event")
            event.set()
    except Exception as e:
        print(e)



from homeassistant.core import HomeAssistant  # type: ignore


async def handle_discovery_feedback(hass: HomeAssistant, info: dict):
    # check if the record already exists if not add it
    if not any(
        device["device_id"] == info["device_id"]
        for device in hass.data["tis_control"]["discovered_devices"]
    ):
        hass.data["tis_control"]["discovered_devices"].append(info)



import asyncio
from TISControlProtocol.shared import ack_events
from typing import Union


class AckCoordinator:
    def __init__(self):
        self.ack_events = ack_events

    def create_ack_event(self, unique_id: Union[str, tuple]) -> asyncio.Event:
        print(f"creating ack event for {unique_id}")
        event = asyncio.Event()
        self.ack_events[unique_id] = event
        return event

    def get_ack_event(self, unique_id: Union[str, tuple]) -> Union[asyncio.Event, None]:
        return self.ack_events.get(unique_id)

    def remove_ack_event(self, unique_id: Union[str, tuple]) -> None:
        if unique_id in self.ack_events:
            del self.ack_events[unique_id]

    # async def create_ack_task(
    #     self,
    #     sender,
    #     packet,
    #     packet_dict=None,
    #     channel_number=None,
    #     attempts=3,
    #     timeout=5.0,
    # ) -> bool:
    #     unique_id = (
    #         (tuple(packet_dict["device_id"])),
    #         packet_dict["operation_code"],
    #         channel_number,
    #     )
    #     event = self.create_ack_event(unique_id)

    #     for attempt in range(attempts):
    #         sender.send_packet(packet)
    #         try:
    #             await asyncio.wait_for(event.wait(), timeout)
    #             self.remove_ack_event(unique_id)
    #             return True
    #         except asyncio.TimeoutError:
    #             print(
    #                 f"ack not received within {timeout} seconds, attempt {attempt + 1}"
    #             )

    #     return False



from homeassistant.core import HomeAssistant  # type: ignore
import logging


class PacketDispatcher:
    def __init__(self, hass: HomeAssistant, OPERATIONS_DICT: dict):
        self.hass = hass
        self.operations_dict = OPERATIONS_DICT

    async def dispatch_packet(self, info):
        try:
            packet_handler = self.operations_dict.get(
                tuple(info["operation_code"]), "unknown operation"
            )
            if packet_handler != "unknown operation":
                await packet_handler(self.hass, info)
            else:
                print(f"unknown operation code: {info['operation_code']}")
        except Exception as e:
            logging.error(f"error in dispatching packet: {e} , {info}")



from TISControlProtocol.Protocols.udp.PacketHandlers.BinaryFeedbackHandler import (
    handle_binary_feedback,  # noqa: F401
)
from TISControlProtocol.Protocols.udp.PacketHandlers.ControlResponseHandler import (
    handle_control_response,  # noqa: F401
)



# Copyright (c) 2014 Adafruit Industries
# Author: Tony DiCola
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import numbers
import time
import numpy as np

import spidev
import RPi.GPIO as GPIO


__version__ = "0.0.4"

BG_SPI_CS_BACK = 0
BG_SPI_CS_FRONT = 1

SPI_CLOCK_HZ = 16000000

ST7789_NOP = 0x00
ST7789_SWRESET = 0x01
ST7789_RDDID = 0x04
ST7789_RDDST = 0x09

ST7789_SLPIN = 0x10
ST7789_SLPOUT = 0x11
ST7789_PTLON = 0x12
ST7789_NORON = 0x13

ST7789_INVOFF = 0x20
ST7789_INVON = 0x21
ST7789_DISPOFF = 0x28
ST7789_DISPON = 0x29

ST7789_CASET = 0x2A
ST7789_RASET = 0x2B
ST7789_RAMWR = 0x2C
ST7789_RAMRD = 0x2E

ST7789_PTLAR = 0x30
ST7789_MADCTL = 0x36
ST7789_COLMOD = 0x3A

ST7789_FRMCTR1 = 0xB1
ST7789_FRMCTR2 = 0xB2
ST7789_FRMCTR3 = 0xB3
ST7789_INVCTR = 0xB4
ST7789_DISSET5 = 0xB6

ST7789_GCTRL = 0xB7
ST7789_GTADJ = 0xB8
ST7789_VCOMS = 0xBB

ST7789_LCMCTRL = 0xC0
ST7789_IDSET = 0xC1
ST7789_VDVVRHEN = 0xC2
ST7789_VRHS = 0xC3
ST7789_VDVS = 0xC4
ST7789_VMCTR1 = 0xC5
ST7789_FRCTRL2 = 0xC6
ST7789_CABCCTRL = 0xC7

ST7789_RDID1 = 0xDA
ST7789_RDID2 = 0xDB
ST7789_RDID3 = 0xDC
ST7789_RDID4 = 0xDD

ST7789_GMCTRP1 = 0xE0
ST7789_GMCTRN1 = 0xE1

ST7789_PWCTR6 = 0xFC


class ST7789(object):
    """Representation of an ST7789 TFT LCD."""

    def __init__(
        self,
        port,
        cs,
        dc,
        backlight=None,
        rst=None,
        width=240,
        height=240,
        rotation=90,
        invert=True,
        spi_speed_hz=4000000,
        offset_left=0,
        offset_top=0,
    ):
        """Create an instance of the display using SPI communication.

        Must provide the GPIO pin number for the D/C pin and the SPI driver.

        Can optionally provide the GPIO pin number for the reset pin as the rst parameter.

        :param port: SPI port number
        :param cs: SPI chip-select number (0 or 1 for BCM
        :param backlight: Pin for controlling backlight
        :param rst: Reset pin for ST7789
        :param width: Width of display connected to ST7789
        :param height: Height of display connected to ST7789
        :param rotation: Rotation of display connected to ST7789
        :param invert: Invert display
        :param spi_speed_hz: SPI speed (in Hz)

        """
        if rotation not in [0, 90, 180, 270]:
            raise ValueError("Invalid rotation {}".format(rotation))

        if width != height and rotation in [90, 270]:
            raise ValueError(
                "Invalid rotation {} for {}x{} resolution".format(
                    rotation, width, height
                )
            )

        GPIO.setwarnings(False)
        GPIO.setmode(GPIO.BCM)

        self._spi = spidev.SpiDev(port, cs)
        self._spi.mode = 0
        self._spi.lsbfirst = False
        self._spi.max_speed_hz = spi_speed_hz

        self._dc = dc
        self._rst = rst
        self._width = width
        self._height = height
        self._rotation = rotation
        self._invert = invert

        self._offset_left = offset_left
        self._offset_top = offset_top

        # Set DC as output.
        GPIO.setup(dc, GPIO.OUT)

        # Setup backlight as output (if provided).
        self._backlight = backlight
        if backlight is not None:
            GPIO.setup(backlight, GPIO.OUT)
            GPIO.output(backlight, GPIO.LOW)
            time.sleep(0.1)
            GPIO.output(backlight, GPIO.HIGH)

        # Setup reset as output (if provided).
        if rst is not None:
            GPIO.setup(self._rst, GPIO.OUT)
            self.reset()
        self._init()

    def send(self, data, is_data=True, chunk_size=4096):
        """Write a byte or array of bytes to the display. Is_data parameter
        controls if byte should be interpreted as display data (True) or command
        data (False).  Chunk_size is an optional size of bytes to write in a
        single SPI transaction, with a default of 4096.
        """
        # Set DC low for command, high for data.
        GPIO.output(self._dc, is_data)
        # Convert scalar argument to list so either can be passed as parameter.
        if isinstance(data, numbers.Number):
            data = [data & 0xFF]
        # Write data a chunk at a time.
        for start in range(0, len(data), chunk_size):
            end = min(start + chunk_size, len(data))
            self._spi.xfer(data[start:end])

    def set_backlight(self, value):
        """Set the backlight on/off."""
        if self._backlight is not None:
            GPIO.output(self._backlight, value)

    @property
    def width(self):
        return (
            self._width
            if self._rotation == 0 or self._rotation == 180
            else self._height
        )

    @property
    def height(self):
        return (
            self._height
            if self._rotation == 0 or self._rotation == 180
            else self._width
        )

    def command(self, data):
        """Write a byte or array of bytes to the display as command data."""
        self.send(data, False)

    def data(self, data):
        """Write a byte or array of bytes to the display as display data."""
        self.send(data, True)

    def reset(self):
        """Reset the display, if reset pin is connected."""
        if self._rst is not None:
            GPIO.output(self._rst, 1)
            time.sleep(0.500)
            GPIO.output(self._rst, 0)
            time.sleep(0.500)
            GPIO.output(self._rst, 1)
            time.sleep(0.500)

    def _init(self):
        # Initialize the display.

        self.command(ST7789_SWRESET)  # Software reset
        time.sleep(0.150)  # delay 150 ms

        self.command(ST7789_MADCTL)
        self.data(0x70)

        self.command(ST7789_FRMCTR2)  # Frame rate ctrl - idle mode
        self.data(0x0C)
        self.data(0x0C)
        self.data(0x00)
        self.data(0x33)
        self.data(0x33)

        self.command(ST7789_COLMOD)
        self.data(0x05)

        self.command(ST7789_GCTRL)
        self.data(0x14)

        self.command(ST7789_VCOMS)
        self.data(0x37)

        self.command(ST7789_LCMCTRL)  # Power control
        self.data(0x2C)

        self.command(ST7789_VDVVRHEN)  # Power control
        self.data(0x01)

        self.command(ST7789_VRHS)  # Power control
        self.data(0x12)

        self.command(ST7789_VDVS)  # Power control
        self.data(0x20)

        self.command(0xD0)
        self.data(0xA4)
        self.data(0xA1)

        self.command(ST7789_FRCTRL2)
        self.data(0x0F)

        self.command(ST7789_GMCTRP1)  # Set Gamma
        self.data(0xD0)
        self.data(0x04)
        self.data(0x0D)
        self.data(0x11)
        self.data(0x13)
        self.data(0x2B)
        self.data(0x3F)
        self.data(0x54)
        self.data(0x4C)
        self.data(0x18)
        self.data(0x0D)
        self.data(0x0B)
        self.data(0x1F)
        self.data(0x23)

        self.command(ST7789_GMCTRN1)  # Set Gamma
        self.data(0xD0)
        self.data(0x04)
        self.data(0x0C)
        self.data(0x11)
        self.data(0x13)
        self.data(0x2C)
        self.data(0x3F)
        self.data(0x44)
        self.data(0x51)
        self.data(0x2F)
        self.data(0x1F)
        self.data(0x1F)
        self.data(0x20)
        self.data(0x23)

        if self._invert:
            self.command(ST7789_INVON)  # Invert display
        else:
            self.command(ST7789_INVOFF)  # Don't invert display

        self.command(ST7789_SLPOUT)

        self.command(ST7789_DISPON)  # Display on
        time.sleep(0.100)  # 100 ms

    def begin(self):
        """Set up the display

        Deprecated. Included in __init__.

        """
        pass

    def set_window(self, x0=0, y0=0, x1=None, y1=None):
        """Set the pixel address window for proceeding drawing commands. x0 and
        x1 should define the minimum and maximum x pixel bounds.  y0 and y1
        should define the minimum and maximum y pixel bound.  If no parameters
        are specified the default will be to update the entire display from 0,0
        to width-1,height-1.
        """
        if x1 is None:
            x1 = self._width - 1

        if y1 is None:
            y1 = self._height - 1

        y0 += self._offset_top
        y1 += self._offset_top

        x0 += self._offset_left
        x1 += self._offset_left

        self.command(ST7789_CASET)  # Column addr set
        self.data(x0 >> 8)
        self.data(x0 & 0xFF)  # XSTART
        self.data(x1 >> 8)
        self.data(x1 & 0xFF)  # XEND
        self.command(ST7789_RASET)  # Row addr set
        self.data(y0 >> 8)
        self.data(y0 & 0xFF)  # YSTART
        self.data(y1 >> 8)
        self.data(y1 & 0xFF)  # YEND
        self.command(ST7789_RAMWR)  # write to RAM

    def display(self, image):
        """Write the provided image to the hardware.

        :param image: Should be RGB format and the same dimensions as the display hardware.

        """
        # Set address bounds to entire display.
        self.set_window()

        # Convert image to 16bit RGB565 format and
        # flatten into bytes.
        pixelbytes = self.image_to_data(image, self._rotation)

        # Write data to hardware.
        for i in range(0, len(pixelbytes), 4096):
            self.data(pixelbytes[i : i + 4096])

    def image_to_data(self, image, rotation=0):
        if not isinstance(image, np.ndarray):
            image = np.array(image.convert("RGB"))

        # Rotate the image
        pb = np.rot90(image, rotation // 90).astype("uint16")

        # Mask and shift the 888 RGB into 565 RGB
        red = (pb[..., [0]] & 0xF8) << 8
        green = (pb[..., [1]] & 0xFC) << 3
        blue = (pb[..., [2]] & 0xF8) >> 3

        # Stick 'em together
        result = red | green | blue

        # Output the raw bytes
        return result.byteswap().tobytes()
